from functools import partial
from typing import Union

import numpy as np
import torch
from anndata import AnnData
from patsy.design_info import DesignMatrix
from torch.types import Device

from .model import guide, model
from .train import SUBSAMPLE, SVILocalHandler
from .utils import get_rna_counts, get_state_loadings, get_states


class scPCA(object):
    """
    scPCA model.

    Parameters
    ----------
    adata: anndata.AnnData
        Anndata object with the single-cell data.
    num_factors: int (default: 15)
        Number of factors to fit.
    layers_key: str or None (default: None)
        Key to extract single-cell count matrix from adata.layers. If layers_key is None,
        scPCA will try to extract the count matrix from the adata.X.
    batch_formula: str or None (default: None)
        R style formula to extract batch information from adata.obs. If batch_formula is None,
        scPCA assumes a single batch. Usually `batch_column - 1`.
    design_formula: str or None (default: None)
        R style formula to construct the design matrix from adata.obs. If design_formula is None,
        scPCA fits a normal PCA.
    subsampling: int (default: 2**12)
        Number of cells to subsample for training. A larger number will result in a more accurate
        computation of the gradients, but will also increase the training time and memory usage.
    device: torch.device (default: torch.device("cuda" if torch.cuda.is_available() else "cpu"))
        Device to run the model on. A GPU is highly recommended.
    model_key: str (default: "scpca")
        Key to store the model in the AnnData object.
    model_kwargs: dict
        Model parameters. See sccca.model.model for more details.
    training_kwargs: dict
        Training parameters. See sccca.handler for more details.
    """

    def __init__(
        self,
        adata: AnnData,
        num_factors: int,
        layers_key: Union[str, None] = None,
        batch_formula: Union[str, None] = None,
        design_formula: Union[str, None] = None,
        subsampling: int = 4096,
        device: Device = torch.device("cuda" if torch.cuda.is_available() else "cpu"),
        model_key: str = "scpca",
        model_kwargs: dict = {
            "β_rna_sd": 0.01,
            "β_rna_mean": 3,
            "intercept": True,
            "batch_beta": False,
            "horseshoe": False,
        },
        training_kwargs: dict = SUBSAMPLE,
    ):
        self.adata = adata
        self.num_factors = num_factors
        self.layers_key = layers_key
        self.batch_formula = batch_formula
        self.design_formula = design_formula
        self.model_key = model_key

        self.subsampling = min([subsampling, adata.shape[0]])
        self.device = device

        # prepare design and batch matrix
        self.batch_matrix = self._get_formula(self.batch_formula)
        self.design_matrix = self._get_formula(self.design_formula)

        #
        self.model_kwargs = model_kwargs
        self.training_kwargs = training_kwargs

        # setup data
        self.data = self._setup_data()
        self.handler = self._setup_handler()

    def _to_torch(self, data):
        """
        Helper method to convert numpy arrays of a dict to torch tensors.
        """
        return {
            k: torch.tensor(v, device=self.device) if isinstance(v, np.ndarray) else v
            for k, v in data.items()
        }

    def _setup_data(self):
        """
        Sets up the data.
        """
        X = get_rna_counts(self.adata, self.layers_key)
        X_size = np.log(X.sum(axis=1, keepdims=True))
        batch = np.asarray(self.batch_matrix).astype(np.float32)
        design = np.asarray(self.design_matrix).astype(np.float32)

        num_genes = X.shape[1]
        num_cells = X.shape[0]
        num_batches = batch.shape[1]
        idx = np.arange(num_cells)

        data = dict(
            X=X,
            X_size=X_size,
            Y=None,
            Y_size=None,
            batch=batch,
            idx=idx,
            num_genes=num_genes,
            num_proteins=None,
            num_batches=num_batches,
            num_cells=num_cells,
            design=design,
        )
        return self._to_torch(data)

    def _setup_handler(self):
        """
        Sets up the handler for training the model.
        """
        train_model = partial(
            model,
            num_factors=self.num_factors,
            subsampling=self.subsampling,
            minibatches=False,
            device=self.device,
            **self.data,
            **self.model_kwargs,
        )

        train_guide = partial(
            guide,
            num_factors=self.num_factors,
            subsampling=self.subsampling,
            minibatches=False,
            device=self.device,
            **self.data,
            **self.model_kwargs,
        )

        idx = self.data.pop("idx")

        predict_model = partial(
            model,
            num_factors=self.num_factors,
            subsampling=0,
            minibatches=True,
            device=self.device,
            **self.data,
            **self.model_kwargs,
        )

        predict_guide = partial(
            guide,
            num_factors=self.num_factors,
            subsampling=0,
            minibatches=True,
            device=self.device,
            **self.data,
            **self.model_kwargs,
        )

        return SVILocalHandler(
            model=train_model,
            guide=train_guide,
            predict_model=predict_model,
            predict_guide=predict_guide,
            idx=idx,
            **self.training_kwargs,
        )

    def fit(self, *args, **kwargs):
        self.handler.fit(*args, **kwargs)

    def to_anndata(self, adata=None, model_key=None, num_samples=25):
        if model_key is None:
            model_key = self.model_key

        if adata is None:
            adata = self.adata

        adata.obsm[f"X_{model_key}"] = self.handler.predict_local_variable(
            "z", num_samples=num_samples
        ).mean(0)
        adata.layers[f"{model_key}_pred_rna"] = self.handler.predict_local_variable(
            "μ_rna", num_samples=num_samples
        ).mean(0)

        adata.varm[f"{model_key}"] = (
            self.handler.predict_global_variable("W_fac", num_samples=num_samples)
            .mean(0)
            .T
        )

        if isinstance(self.design_matrix, DesignMatrix):
            adata.uns[f"{model_key}"] = {"design": get_states(self.design_matrix)}

        state_loadings = get_state_loadings(adata, model_key)
        adata.uns[f"{model_key}"]["states"] = state_loadings
        adata.uns[f"{model_key}"]["α_rna"] = self.handler.predict_global_variable(
            "α_rna"
        ).mean(0)



from typing import Union

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.colors import LogNorm
from scipy.stats import variation
from sklearn.metrics import mean_squared_error

from ..utils import get_protein_counts, get_rna_counts


def _extract_counts(adata, layers_key, protein_obsm_key):
    if protein_obsm_key is not None:
        counts = get_protein_counts(adata, protein_obsm_key)
    else:
        counts = get_rna_counts(adata, layers_key)

    return counts


def disp(
    adata,
    model_key: str = "scpca",
    layers_key: Union[str, None] = None,
    protein_obsm_key: Union[str, None] = None,
    cmap: str = "viridis",
    ax: plt.Axes = None,
) -> plt.Axes:
    """
    Plot the fitted dispersion against the coefficient of variation RNA or
    protein counts.

    Parameters
    ----------
    adata: AnnData
        AnnData object.
    model_key: str, optional (default: "scpca")
        Key for the fitted model.
    layers_key: str, optional (default: None)
        If `layers_key` is None, then the raw counts are extracted from `adata.X`.
        Otherwise, the counts are extracted from `adata.layers[layers_key]`.
    protein_obsm_key: str, optional (default: None)
        Key for protein counts in `adata.obsm`. Providing `protein_obsm_key`
        overrides `layers_key`, i.e. protein counts are plotted.
    cmap: str, optional (default: "viridis")
        Colormap for the scatterplot. Color represents the mean of the counts.
    ax: matplotlib.axes.Axes, optional (default: None)
        Axes to plot on. If None, then a new figure is created.

    Returns
    -------
    ax: matplotlib.axes.Axes
    """
    # Extract counts
    counts = _extract_counts(adata, layers_key, protein_obsm_key)
    posterior_key = "α_prot" if protein_obsm_key is not None else "α_rna"

    if ax is None:
        plt.scatter(
            adata.uns[model_key]["posterior"][posterior_key],
            variation(counts, axis=0),
            c=counts.mean(0),
            cmap="viridis",
            norm=LogNorm(),
        )
        ax = plt.gca()
    else:
        ax.scatter(
            adata.uns[model_key]["posterior"][posterior_key],
            variation(counts, axis=0),
            c=counts.mean(0),
            cmap="viridis",
            norm=LogNorm(),
        )

    plt.colorbar()
    # ax.plot(np.linspace(1, 60), np.linspace(1, 60), color='C1')
    ax.set_yscale("log")
    ax.set_xscale("log")
    ax.set_xlabel(r"$\alpha$")
    ax.set_ylabel("CV")
    return ax


def qc_hist(
    adata,
    model_key: str = "scpca",
    layers_key: Union[str, None] = None,
    protein_obsm_key: Union[str, None] = None,
    cmap: str = "viridis",
    ax: plt.Axes = None,
) -> plt.Axes:
    """
    Plots a 2D histogram of the predicted counts against the true counts.

    Parameters
    ----------
    adata: AnnData
        AnnData object.
    model_key: str, optional (default: "scpca")
        Key for the fitted model.
    layers_key: str, optional (default: None)
        If `layers_key` is None, then the raw counts are extracted from `adata.X`.
        Otherwise, the counts are extracted from `adata.layers[layers_key]`.
    protein_obsm_key: str, optional (default: None)
        Key for protein counts in `adata.obsm`. Providing `protein_obsm_key`
        overrides `layers_key`, i.e. protein counts are plotted.
    cmap: str, optional (default: "viridis")
        Colormap for the scatterplot. Color represents the mean of the counts.
    ax: matplotlib.axes.Axes, optional (default: None)
        Axes to plot on. If None, then a new figure is created.

    Returns
    -------
    ax: matplotlib.axes.Axes
    """
    # Extract counts
    counts = _extract_counts(adata, layers_key, protein_obsm_key)
    posterior_key = "μ_prot" if protein_obsm_key is not None else "μ_rna"
    predicted_counts = adata.uns[model_key]["posterior"][posterior_key]

    if ax is None:
        plt.hist2d(
            np.log10(counts.reshape(-1) + 1),
            np.log10(predicted_counts.reshape(-1) + 1),
            bins=50,
            norm=LogNorm(),
            cmap=cmap,
        )
        ax = plt.gca()
    else:
        ax.hist2d(
            np.log10(counts.reshape(-1) + 1),
            np.log10(predicted_counts.reshape(-1) + 1),
            bins=50,
            norm=LogNorm(),
            cmap=cmap,
        )
    plt.colorbar()
    max_val = np.max([*ax.get_xlim(), *ax.get_ylim()])
    min_val = np.min([*ax.get_xlim(), *ax.get_ylim()])
    # print(max_val)
    ax.set_xlim([min_val, max_val])
    ax.set_ylim([min_val, max_val])
    ax.plot(
        np.linspace(min_val, max_val),
        np.linspace(min_val, max_val),
        color="w",
        linewidth=2,
    )
    ax.set_aspect("equal")
    ax.set_ylabel(r"Predicted count ($\log_{10}(x+1)$ scaled)", fontsize=12)
    ax.set_xlabel(r"True count ($\log_{10}(x+1)$ scaled)", fontsize=12)
    rmse = mean_squared_error(counts, predicted_counts)
    ax.set_title(f"RMSE {rmse:.2f}")
    return ax



from typing import Union

import numpy as np
import torch
from anndata import AnnData
from torch.types import Device

from .pca import scPCA
from .train import SUBSAMPLE
from .utils import get_protein_counts, get_rna_counts, get_states


class scCCA(scPCA):
    """
    scCCA model.

    Parameters
    ----------
    adata: anndata.AnnData
        Anndata object with the single-cell data.
    num_factors: int
        Number of factors to fit.
    protein_obsm_key: str or None (default: None)
        Key to extract single-cell protein matrix from `adata.obsm`.
    layers_key: str or None (default: None)
        Key to extract single-cell count matrix from adata.layers. If layers_key is None,
        scPCA will try to extract the count matrix from the adata.X.
    batch_formula: str or None (default: None)
        R style formula to extract batch information from adata.obs. If batch_formula is None,
        scPCA assumes a single batch. Usually `batch_column - 1`.
    design_formula: str or None (default: None)
        R style formula to construct the design matrix from adata.obs. If design_formula is None,
        scPCA fits a normal PCA.
    subsampling: int (default: 4096)
        Number of cells to subsample for training. A larger number will result in a more accurate
        computation of the gradients, but will also increase the training time and memory usage.
    device: torch.device (default: torch.device("cuda") if a GPU is available)
        Device to run the model on. A GPU is highly recommended.
    model_key: str (default: "scpca")
        Key to store the model in the AnnData object.
    model_kwargs: dict
        Model parameters. See sccca.model.model for more details.
    training_kwargs: dict
        Training parameters. See sccca.handler for more details.
    """

    def __init__(
        self,
        adata: AnnData,
        num_factors: int,
        protein_obsm_key: str,
        layers_key: Union[str, None] = None,
        batch_formula: Union[str, None] = None,
        design_formula: Union[str, None] = None,
        subsampling: int = 4096,
        device: Device = torch.device("cuda" if torch.cuda.is_available() else "cpu"),
        model_key: str = "sccca",
        model_kwargs: dict = {
            "β_rna_sd": 0.01,
            "β_rna_mean": 3,
            "intercept": True,
            "batch_beta": False,
            "horseshoe": False,
        },
        training_kwargs: dict = SUBSAMPLE,
    ):
        self.protein_obsm_key = protein_obsm_key

        super().__init__(
            adata=adata,
            num_factors=num_factors,
            layers_key=layers_key,
            batch_formula=batch_formula,
            design_formula=design_formula,
            subsampling=subsampling,
            device=device,
            model_key=model_key,
            model_kwargs=model_kwargs,
            training_kwargs=training_kwargs,
        )

    def _setup_data(self):
        """
        Sets up the data.
        """
        X = get_rna_counts(self.adata, self.layers_key)
        Y = get_protein_counts(self.adata, self.protein_obsm_key)
        X_size = np.log(X.sum(axis=1, keepdims=True))
        Y_size = np.log(Y.sum(axis=1, keepdims=True))
        batch = np.asarray(self.batch_matrix).astype(np.float32)
        design = np.asarray(self.design_matrix).astype(np.float32)

        num_genes = X.shape[1]
        num_cells = X.shape[0]
        num_batches = batch.shape[1]
        num_proteins = Y.shape[1]
        idx = np.arange(num_cells)

        data = dict(
            X=X,
            X_size=X_size,
            Y=Y,
            Y_size=Y_size,
            batch=batch,
            idx=idx,
            num_genes=num_genes,
            num_proteins=num_proteins,
            num_batches=num_batches,
            num_cells=num_cells,
            design=design,
        )
        return self._to_torch(data)

    def to_anndata(self, adata=None, model_key=None, num_samples=25):
        if model_key is None:
            model_key = self.model_key

        if adata is None:
            adata = self.adata

        # set up unstructured metadata
        adata.uns[f"{model_key}"] = {}
        adata.uns[f"{model_key}"]["posterior"] = {
            "μ_rna": self.handler.predict_local_variable(
                "μ_rna", num_samples=num_samples
            ).mean(0),
            "μ_prot": self.handler.predict_local_variable(
                "μ_prot", num_samples=num_samples
            ).mean(0),
            "α_rna": self.handler.predict_global_variable(
                "α_rna", num_samples=num_samples
            ).mean(0),
            "α_prot": self.handler.predict_global_variable(
                "α_prot", num_samples=num_samples
            ).mean(0),
            "V_prot": self.handler.predict_global_variable(
                "V_fac", num_samples=num_samples
            )
            .mean(0)
            .T,
        }

        adata.uns[f"{model_key}"]["design"] = get_states(self.design_matrix)
        # adata.uns[f"{model_key}"]["batch"] = get_states(self.batch)
        adata.uns[f"{model_key}"]["model"] = self.model_kwargs

        adata.obsm[f"X_{model_key}"] = self.handler.predict_local_variable(
            "z", num_samples=num_samples
        ).mean(0)

        adata.varm[f"{model_key}"] = (
            self.handler.predict_global_variable("W_fac", num_samples=num_samples)
            .mean(0)
            .T
        )



from collections import defaultdict

import numpy as np
import pyro
import torch
from pyro.infer import SVI, Predictive, Trace_ELBO
from tqdm import tqdm


class SVIBaseHandler:
    """
    Helper object that abstracts some of numpyros complexities. Inspired
    by an implementation of Florian Wilhelm.
    :param model: A numpyro model.
    :param guide: A numpyro guide.
    :param loss: Loss function, defaults to Trace_ELBO.
    :param lr: Learning rate, defaults to 0.01.
    :param rng_key: Random seed, defaults to 254.
    :param num_epochs: Number of epochs to train the model, defaults to 5000.
    :param num_samples: Number of posterior samples.
    :param log_func: Logging function, defaults to print.
    :param log_freq: Frequency of logging, defaults to 0 (no logging).
    :param to_numpy: Convert the posterior distribution to numpy array(s),
        defaults to True.
    """

    def __init__(
        self,
        model,
        guide,
        loss: Trace_ELBO = pyro.infer.TraceMeanField_ELBO,
        optimizer=torch.optim.Adam,
        scheduler=pyro.optim.ReduceLROnPlateau,
        seed=None,
        num_epochs: int = 30000,
        log_freq: int = 10,
        checkpoint_freq: int = -1,
        to_numpy: bool = True,
        optimizer_kwargs: dict = {"lr": 1e-2},
        scheduler_kwargs: dict = {"factor": 0.99},
        loss_kwargs: dict = {"num_particles": 1},
    ):
        pyro.clear_param_store()
        if seed is not None:
            pyro.set_rng_seed(seed)
        self.model = model
        self.guide = guide
        self.loss = loss(**loss_kwargs)
        self.scheduler = False if scheduler is None else True
        self.seed = seed

        if self.scheduler:
            self.optimizer = scheduler(
                {
                    "optimizer": optimizer,
                    "optim_args": optimizer_kwargs,
                    **scheduler_kwargs,
                }
            )
        else:
            self.optimizer = optimizer(optimizer_kwargs)

        self.svi = self._init_svi()
        self.log_freq = log_freq
        self.checkpoint_freq = checkpoint_freq
        self.learning_rates = defaultdict(list)
        self.gradient_norms = defaultdict(list)

        self.checkpoints = {}
        self.num_epochs = num_epochs

        self.loss = None
        self.to_numpy = to_numpy
        self.best_elbo = None
        self.steps = 0

    def _update_state(self, loss):
        self.loss = loss if self.loss is None else np.concatenate([self.loss, loss])

    def _to_numpy(self, posterior):
        return {k: v.detach().cpu().numpy() for k, v in posterior.items()}

    def _init_svi(self):
        """
        Initialises the SVI.
        """
        return SVI(self.model, self.guide, self.optimizer, loss=self.loss)

    def _set_lr(self, lr):
        # this works for CLIPPED adam
        # checl if optim_objs is presetn
        if hasattr(self.optimizer, "optim_objs"):
            for opt in self.optimizer.optim_objs.values():
                for group in opt.param_groups:
                    group["lr"] = lr
        else:
            pass
            # for group in self.optimizer.optim_objs.values():
            #     group['lr'] = lr

    def _get_learning_rate(self):
        """
        Extracts the learning rate from the first parameter.
        TODO: return a dict(lr: [params])
        """
        for name, param in self.optimizer.get_state().items():
            if "optimizer" in param:
                lr = param["optimizer"]["param_groups"][0]["lr"]
            else:
                lr = param["param_groups"][0]["lr"]
            break
        return lr

    def _track_learning_rate(self):
        """
        Tracks the learning rate during training.
        """
        for name, param in self.optimizer.get_state().items():
            if "optimizer" in param:
                self.learning_rates[name].append(
                    param["optimizer"]["param_groups"][0]["lr"]
                )
            else:
                self.learning_rates[name].append(param["param_groups"][0]["lr"])

    def _track_gradient_norms(self):
        # Register hooks to monitor gradient norms.

        for name, value in pyro.get_param_store().named_parameters():
            value.register_hook(
                lambda g, name=name: self.gradient_norms[name].append(g.norm().item())
            )

    def _fit(self, *args, **kwargs):
        losses = []
        pbar = tqdm(range(self.steps, self.steps + self.num_epochs))
        failure = False

        previous_elbo = 0
        best_elbo = np.inf if self.best_elbo is None else self.best_elbo
        delta = 0

        try:
            for i in pbar:
                current_elbo = self.svi.step(*args, **kwargs)
                losses.append(current_elbo)

                if i == 0:
                    self._track_gradient_norms()
                    best_elbo = current_elbo
                else:
                    improvement = best_elbo - current_elbo
                    if improvement > 0:
                        best_elbo = current_elbo

                if i % self.log_freq == 0:
                    lr = self._get_learning_rate()
                    pbar.set_description(
                        f"Epoch: {i} | lr: {lr:.2E} | ELBO: {current_elbo:.0f} | Δ_{self.log_freq}: {delta:.2f} | Best: {best_elbo:.0f}"
                    )
                    if i > 0:
                        delta = previous_elbo - current_elbo
                    previous_elbo = current_elbo

                if self.checkpoint_freq > 0 and i % self.checkpoint_freq == 0:
                    self.checkpoints[i] = pyro.get_param_store().get_state()

                if self.scheduler:
                    if issubclass(
                        self.optimizer.pt_scheduler_constructor,
                        torch.optim.lr_scheduler.ReduceLROnPlateau,
                    ):
                        self.optimizer.step(current_elbo)
                    else:
                        self.optimizer.step()

        except KeyboardInterrupt:
            # TODO: use warning
            print("Stoping training ...")
            failure = True

        if failure:
            self.steps += i
        else:
            self.steps += self.num_epochs

        # update max elbo
        self.best_elbo = best_elbo

        return losses

    def fit(self, *args, **kwargs):
        self.num_epochs = kwargs.pop("num_epochs", self.num_epochs)
        lr = kwargs.pop("lr", None)
        if lr is not None:
            self._set_lr(lr)

        loss = self._fit(*args, **kwargs)
        self._update_state(loss)
        self.params = pyro.get_param_store().get_state()

    def predict(self, return_sites, num_samples=25, *args, **kwargs):
        if self.params is not None:
            pyro.clear_param_store()
            pyro.get_param_store().set_state(self.params)
        """kwargs -> Predictive, args -> predictive"""
        predictive = Predictive(
            self.model,
            guide=self.guide,
            num_samples=num_samples,
            return_sites=return_sites,
        )
        posterior = predictive(*args, **kwargs)
        self.posterior = self._to_numpy(posterior) if self.to_numpy else posterior
        torch.cuda.empty_cache()



import pyro
import torch
from pyro import deterministic, param, plate, sample
from pyro.distributions import (
    Exponential,
    Gamma,
    GammaPoisson,
    HalfCauchy,
    InverseGamma,
    Normal,
    TransformedDistribution,
)
from pyro.distributions.constraints import less_than, positive
from pyro.distributions.transforms import ExpTransform
from torch import einsum, exp, ones, tensor, zeros


def model(
    X,
    Y,
    X_size,
    Y_size,
    batch,
    idx,
    num_genes,
    num_proteins,
    num_batches,
    num_cells,
    num_factors,
    design,
    device,
    subsampling=0,
    β_rna_mean=3.,
    β_rna_sd=.1,
    W_fac_sd=1.,
    z_sd=0.1,
    horseshoe=False,
    batch_beta=False,
    intercept=True,
    constrain_alpha=False,
    fix_alpha_rna=None,
    minibatches=False,
    *args,
    **kwargs
):

    gene_plate = plate("genes", num_genes)
    
    if Y is not None:
        protein_plate = plate("proteins", num_proteins)
    
    # cell_plate = plate("cells", num_cells)
    batch_plate = plate("batches", num_batches)
    factor_plate = plate("factor", num_factors)
    # num groups from the design matrix
    num_groups = design.shape[1]
    group_plate = plate('groups', num_groups)
    
    if subsampling > 0:
        cell_plate = plate("cells", num_cells, subsample_size=subsampling)
    else:
        cell_plate = plate("cells", num_cells, subsample=idx)

    # factor matrices
    if horseshoe:
        with pyro.plate('modality', 2):
            tau = pyro.sample('tau', HalfCauchy(ones(1, device=device)))                
    
    with group_plate:

        # W_fac = sample("W_fac", Normal(zeros(num_genes, device=device), ones(num_genes, device=device)).to_event(2))
        W_fac = sample("W_fac", Normal(zeros((num_factors, num_genes), device=device), W_fac_sd * ones((num_factors, num_genes), device=device)).to_event(2))
        # scale additive factors
        # W_fac = W_fac * torch.cat([torch.ones((1, num_factors, num_genes), device=device), W_fac_sd * torch.ones((num_groups-1, num_factors, num_genes), device=device)], 0)
        # torch.ones((num_groups, num_factors, num_genes), device=device)
        # print(W_fac.shape)11
        if Y is not None:
            V_fac = sample("V_fac", Normal(zeros((num_factors, num_proteins), device=device), ones((num_factors, num_proteins), device=device)).to_event(2))

        if horseshoe: # turns on sparsity priors
            W_del = sample('W_del', HalfCauchy(ones(1, device=device)))
            W_lam = sample('W_lam', HalfCauchy(ones(num_genes, device=device)).to_event(1))
            W_c = sample("W_c", InverseGamma(0.5 * ones(num_genes, device=device), 0.5 * ones(num_genes, device=device)).to_event(1))

            # print(W_tau.shape, W_lam.shape, W_fac.shape)
            W_gamma = (tau[0] * W_del.reshape(-1, 1) * W_lam)
            W_fac = deterministic("W_horse", W_fac * (torch.sqrt(W_c) * W_gamma) / torch.sqrt(W_c + W_gamma**2))

            V_del = sample('V_del', HalfCauchy(ones(1, device=device)))
            V_lam = sample('V_lam', HalfCauchy(ones(num_proteins, device=device)).to_event(1))            
            V_c = sample("V_c", InverseGamma(0.5 * ones(num_proteins, device=device), 0.5 * ones(num_proteins, device=device)).to_event(1))

            V_gamma = (tau[1] * V_del.reshape(-1, 1) * V_lam)
            V_fac = deterministic("V_horse", V_fac * (torch.sqrt(V_c) * V_gamma) / torch.sqrt(V_c + V_gamma**2))

    # if design is not None:
    #     with group_plate:    
    #         W_grp = sample("W_grp", Normal(zeros((num_factors, num_genes), device=device), ones((num_factors, num_genes), device=device)).to_event(2))
    #         # print('W_grp', W_grp.shape)
    #         # base = zeros((1, num_factors, num_genes), device=device)
    #             # print(Normal(zeros(num_genes, device=device), ones(num_genes, device=device)).to_event(1).batch_shape)
    #         W_grp_cat = torch.cat([W_fac.unsqueeze(0), W_grp], 0)
            
    if intercept:
        with batch_plate:
            W_add = sample("W_add", Normal(zeros(num_genes, device=device), ones(num_genes, device=device)).to_event(1))
            
            if Y is not None:
                V_add = sample("V_add", Normal(zeros(num_proteins, device=device), ones(num_proteins, device=device)).to_event(1))
    else:
        W_add = torch.zeros((num_batches, num_genes), device=device)
        
        if Y is not None:
            V_add = torch.zeros((num_batches, num_proteins), device=device)
        
            
    β_rna_conc = β_rna_mean**2/β_rna_sd**2,
    β_rna_rate = β_rna_mean/β_rna_sd**2,
    
    if batch_beta:
        with batch_plate:
            β_rna = sample("β_rna", Gamma(tensor(β_rna_conc, device=device), tensor(β_rna_rate, device=device)))
            
            if Y is not None:
                β_prot = sample("β_prot", Gamma(tensor(9.0, device=device), tensor(3.0, device=device)))
        
        with gene_plate:
            α_rna_inv = sample("α_rna_inv", Exponential(β_rna).to_event(1))
            α_rna = deterministic("α_rna", (1 / α_rna_inv).T)
        
        # print('model', α_rna_inv.shape)
        if Y is not None:
            with protein_plate:
                α_prot_inv = sample("α_prot_inv", Exponential(β_prot).to_event(1))
                α_prot = deterministic("α_prot", (1 / α_prot_inv).T)

    else:
        β_rna = sample("β_rna", Gamma(tensor(β_rna_conc, device=device), tensor(β_rna_rate, device=device)))
        if Y is not None:
            β_prot = sample("β_prot", Gamma(tensor(9.0, device=device), tensor(3.0, device=device)))

        with gene_plate:
            α_rna_inv = sample("α_rna_inv", Exponential(β_rna))
            α_rna = deterministic("α_rna", (1 / α_rna_inv).T)
            
            # print('model', α_rna_inv.shape)
            
        if Y is not None:
            with protein_plate:
                α_prot_inv = sample("α_prot_inv", Exponential(β_prot))
                α_prot = deterministic("α_prot", (1 / α_prot_inv).T)

            
    if subsampling > 0 or minibatches:
        with cell_plate as ind:
            z = sample("z", Normal(zeros(num_factors, device=device), z_sd * ones(num_factors, device=device),).to_event(1),)

            W_rna = (design[ind].T.unsqueeze(-1) * einsum("cf,bfp->bcp", z,  W_fac)).sum(0)
            # print(batch[ind].shape, W_add.shape)
            intercept_rna = batch[ind] @ W_add

            μ_rna = deterministic("μ_rna", exp(X_size[ind] + intercept_rna + W_rna))
            
            if Y is not None:
                V_prot = (design[ind].T.unsqueeze(-1) * einsum("cf,bfp->bcp", z,  V_fac)).sum(0)
                intercept_prot = batch[ind] @ V_add                
                
                μ_prot = deterministic("μ_prot", exp(Y_size[ind] + intercept_prot + V_prot))
            
            if batch_beta:
                α_rna_bat = batch[ind] @ α_rna
                if Y is not None:
                    α_prot_bat = batch[ind] @ α_prot
            else:
                α_rna_bat = α_rna
                
                if Y is not None:
                    α_prot_bat = α_prot
            
            sample("rna", GammaPoisson(α_rna_bat, α_rna_bat / μ_rna).to_event(1), obs=X[ind])
            if Y is not None:
                sample("prot", GammaPoisson(α_prot_bat, α_prot_bat / μ_prot).to_event(1), obs=Y[ind])
    else:
        with cell_plate:
            z = sample("z", Normal(zeros(num_factors, device=device),0.1 * ones(num_factors, device=device),).to_event(1),)

            μ_rna = deterministic("μ_rna", exp((batch.T.unsqueeze(-1) * (X_size.unsqueeze(0) + W_add.unsqueeze(1) + einsum("cf,bfp->bcp", z, W_fac.unsqueeze(0)))).sum(0)))
            
            if Y is not None:
                μ_prot = deterministic("μ_prot", exp((batch.T.unsqueeze(-1) * (Y_size.unsqueeze(0) + V_add.unsqueeze(1) + einsum("cf,bfp->bcp", z, V_fac.unsqueeze(0)))).sum(0)))
            
            if batch_beta:
                α_rna_bat = batch @ α_rna
                if Y is not None:
                    α_prot_bat = batch @ α_prot
            else:
                α_rna_bat = α_rna
                if Y is not None:
                    α_prot_bat = α_prot
            
            sample("rna", GammaPoisson(α_rna_bat, α_rna_bat / μ_rna).to_event(1), obs=X)
            
            if Y is not None:
                sample("prot", GammaPoisson(α_prot_bat, α_prot_bat / μ_prot).to_event(1), obs=Y)
                

def guide(
    X,
    Y,
    X_size,
    Y_size,
    batch,
    idx,
    num_genes,
    num_proteins,
    num_batches,
    num_cells,
    num_factors,
    design,
    device,
    subsampling=0,
    β_rna_mean=3.,
    β_rna_sd=.1,
    W_fac_sd=1.,
    z_sd=1.,
    horseshoe=False,
    batch_beta=False,
    intercept=True,
    constrain_alpha=False,
    fix_alpha_rna=None,
    minibatches=False,
    *args,
    **kwargs
):

    gene_plate = plate("genes", num_genes)
    if Y is not None:
        protein_plate = plate("proteins", num_proteins)
    batch_plate = plate("batches", num_batches)
    # cell_plate = plate("cells", num_cells)
    factor_plate = plate("factor", num_factors)
    
    # design matrix
    num_groups = design.shape[1] 
    group_plate = plate('groups', num_groups)
    
    
    if subsampling > 0:
        cell_plate = plate("cells", num_cells, subsample_size=subsampling)
    else:
        cell_plate = plate("cells", num_cells, subsample=idx)
        
    W_fac_loc = param("W_fac_loc", zeros((num_groups, num_factors, num_genes), device=device))
    W_fac_scale = param( "W_fac_scale", 0.1 * ones((num_groups, num_factors, num_genes), device=device), constraint=positive)
    
    if Y is not None:
        V_fac_loc = param("V_fac_loc", zeros((num_groups, num_factors, num_proteins), device=device))
        V_fac_scale = param("V_fac_scale", 0.1 * ones((num_groups, num_factors, num_proteins), device=device), constraint=positive)
        
    if horseshoe:
        tau_loc = param("tau_loc", zeros(2, device=device))
        tau_scale = param( "tau_scale", 0.1 * ones(2, device=device), constraint=positive)        
        
        with pyro.plate('modality', 2):
            sample("tau", TransformedDistribution(Normal(tau_loc, tau_scale), ExpTransform()))
            
        W_del_loc = param("W_del_loc", zeros(num_factors, device=device))
        W_del_scale = param( "W_del_scale", 0.1 * ones(num_factors, device=device), constraint=positive)        
        
        W_lam_loc = param("W_lam_loc", zeros((num_factors, num_genes), device=device))
        W_lam_scale = param( "W_lam_scale", 0.1 * ones((num_factors, num_genes), device=device), constraint=positive)
        
        W_c_loc = param("W_c_loc", zeros((num_factors, num_genes), device=device))
        W_c_scale = param( "W_c_scale", 0.1 * ones((num_factors, num_genes), device=device), constraint=positive)
        
        V_del_loc = param("V_del_loc", zeros(num_factors, device=device))
        V_del_scale = param("V_del_scale", 0.1 * ones(num_factors, device=device), constraint=positive)        
        
        V_lam_loc = param("V_lam_loc", zeros((num_factors, num_proteins), device=device))
        V_lam_scale = param( "V_lam_scale", 0.1 * ones((num_factors, num_proteins), device=device), constraint=positive)        
        
        V_c_loc = param("V_c_loc", zeros((num_factors, num_proteins), device=device))
        V_c_scale = param( "V_c_scale", 0.1 * ones((num_factors, num_proteins), device=device), constraint=positive)            
    
    with group_plate:
        sample("W_fac", Normal(W_fac_loc, W_fac_scale).to_event(2))

        if Y is not None:
            sample("V_fac", Normal(V_fac_loc, V_fac_scale).to_event(2))

        if horseshoe:
            sample("W_del", TransformedDistribution(Normal(W_del_loc, W_del_scale), ExpTransform()))
            sample("W_lam", TransformedDistribution(Normal(W_lam_loc, W_lam_scale), ExpTransform()).to_event(1))
            sample("W_c", TransformedDistribution(Normal(W_c_loc, W_c_scale), ExpTransform()).to_event(1))

            sample("V_del", TransformedDistribution(Normal(V_del_loc, V_del_scale), ExpTransform()))
            sample("V_lam", TransformedDistribution(Normal(V_lam_loc, V_lam_scale), ExpTransform()).to_event(1))                    
            sample("V_c", TransformedDistribution(Normal(V_c_loc, V_c_scale), ExpTransform()).to_event(1))                    
    
#     if design is not None:
#         W_grp_loc = param("W_grp_loc", zeros((num_groups, num_factors, num_genes), device=device))
#         W_grp_scale = param( "W_grp_scale", 0.1 * ones((num_groups, num_factors, num_genes), device=device), constraint=positive)

#         with group_plate:
#             sample("W_grp", Normal(W_grp_loc, W_grp_scale).to_event(2))
    
    # intercept terms
    if intercept:
        W_add_loc = param("W_add_loc", zeros((num_batches, num_genes), device=device),)
        W_add_scale = param("W_add_scale", 0.1 * ones((num_batches, num_genes), device=device), constraint=positive)
        
        if Y is not None:
            V_add_loc = param( "V_add_loc", zeros((num_batches, num_proteins), device=device),)
            V_add_scale = param("V_add_scale", 0.1 * ones((num_batches, num_proteins), device=device), constraint=positive)

        with batch_plate:
            sample("W_add", Normal(W_add_loc, W_add_scale).to_event(1))
            
            if Y is not None:
                sample("V_add", Normal(V_add_loc, V_add_scale).to_event(1))
            
    # account for batches in beta
    if batch_beta:
        β_rna_loc = param("β_rna_loc", zeros(num_batches, device=device))
        β_rna_scale = param("β_rna_scale", ones(num_batches, device=device), constraint=positive)
        
        if Y is not None:
            β_prot_loc = param("β_prot_loc", zeros(num_batches, device=device))
            β_prot_scale = param("β_prot_scale", ones(num_batches, device=device), constraint=positive)
        
        with batch_plate:
            sample("β_rna", TransformedDistribution(Normal(β_rna_loc, β_rna_scale), ExpTransform()))
            
            if Y is not None:
                sample("β_prot", TransformedDistribution(Normal(β_prot_loc, β_prot_scale), ExpTransform()))
    else:
        β_rna_loc = param("β_rna_loc", zeros(1, device=device))
        β_rna_scale = param("β_rna_scale", ones(1, device=device), constraint=positive)
        
        if Y is not None:
            β_prot_loc = param("β_prot_loc", zeros(1, device=device))
            β_prot_scale = param("β_prot_scale", ones(1, device=device), constraint=positive)        
        
        sample("β_rna", TransformedDistribution(Normal(β_rna_loc, β_rna_scale), ExpTransform()))
        
        if Y is not None:
            sample("β_prot", TransformedDistribution(Normal(β_prot_loc, β_prot_scale), ExpTransform()))
        # print('guide', β_rna_loc.shape)            
        
    if batch_beta:
        if constrain_alpha:
            α_rna_loc = param("α_rna_loc", zeros((num_genes, num_batches), device=device), constraint=less_than(6.0))
        else:
            α_rna_loc = param("α_rna_loc", zeros((num_genes, num_batches), device=device))
            
        α_rna_scale = param("α_rna_scale", 0.1 * ones((num_genes, num_batches), device=device), constraint=positive)
        
        if Y is not None:
            α_prot_loc = param("α_prot_loc", zeros((num_proteins, num_batches), device=device))
            α_prot_scale = param("α_prot_scale", 0.1 * ones((num_proteins, num_batches), device=device), constraint=positive)

        with gene_plate:
            sample("α_rna_inv", TransformedDistribution(Normal(α_rna_loc, α_rna_scale), ExpTransform()).to_event(1))
        
        if Y is not None:
            with protein_plate:
                sample("α_prot_inv", TransformedDistribution(Normal(α_prot_loc, α_prot_scale), ExpTransform()).to_event(1))
    else:
        if constrain_alpha:
            α_rna_loc = param("α_rna_loc", zeros((num_genes), device=device), constraint=less_than(6.0))
        else:
            α_rna_loc = param("α_rna_loc", zeros((num_genes), device=device))
            
        α_rna_scale = param("α_rna_scale", 0.1 * ones((num_genes), device=device), constraint=positive)
        
        if Y is not None:
            α_prot_loc = param("α_prot_loc", zeros((num_proteins), device=device))
            α_prot_scale = param("α_prot_scale", 0.1 * ones((num_proteins), device=device), constraint=positive)

        with gene_plate:
            sample("α_rna_inv", TransformedDistribution(Normal(α_rna_loc, α_rna_scale), ExpTransform()))
        
        if Y is not None:    
            with protein_plate:
                sample("α_prot_inv", TransformedDistribution(Normal(α_prot_loc, α_prot_scale), ExpTransform()))    
    
    z_loc = param("z_loc", zeros((num_cells, num_factors), device=device))
    z_scale = param("z_scale", 0.1 * ones((num_cells, num_factors), device=device), constraint=positive)

    with cell_plate as ind:
        sample("z", Normal(z_loc[ind], z_scale[ind]).to_event(1))    



import numpy as np
import pyro
import torch
from pyro.infer import Predictive, Trace_ELBO
from tqdm import tqdm

from .handler import SVIBaseHandler


class SVILocalHandler(SVIBaseHandler):
    """
    Extend SVIBaseHandler to enable to use a separate model and guide
    for prediction. Assumes theat model and guide accept an idx argument
    that is an torch array of indices.
    """

    def __init__(
        self,
        model,
        guide,
        loss: Trace_ELBO = pyro.infer.TraceMeanField_ELBO,
        optimizer=torch.optim.Adam,
        scheduler=pyro.optim.ReduceLROnPlateau,
        seed=None,
        num_epochs: int = 30000,
        log_freq: int = 10,
        checkpoint_freq: int = 500,
        to_numpy: bool = True,
        optimizer_kwargs: dict = {"lr": 1e-2},
        scheduler_kwargs: dict = {"factor": 0.99},
        loss_kwargs: dict = {"num_particles": 1},
        predict_model=None,
        predict_guide=None,
        idx: torch.Tensor = None,
    ):
        super().__init__(
            model=model,
            guide=guide,
            loss=loss,
            optimizer=optimizer,
            scheduler=scheduler,
            seed=seed,
            num_epochs=num_epochs,
            log_freq=log_freq,
            checkpoint_freq=checkpoint_freq,
            to_numpy=to_numpy,
            optimizer_kwargs=optimizer_kwargs,
            scheduler_kwargs=scheduler_kwargs,
            loss_kwargs=loss_kwargs,
        )
        self.predict_model = predict_model
        self.predict_guide = predict_guide
        self.idx = idx

    def predict(self, return_sites, num_samples=25, *args, **kwargs):
        if self.params is not None:
            pyro.clear_param_store()
            pyro.get_param_store().set_state(self.params)

        predictive = Predictive(
            self.predict_model,
            guide=self.predict_guide,
            num_samples=num_samples,
            return_sites=return_sites,
        )

        posterior = predictive(*args, **kwargs)
        self.posterior = self._to_numpy(posterior) if self.to_numpy else posterior
        torch.cuda.empty_cache()

    def predict_global_variable(self, var: str, num_samples: int = 25):
        """
        Sample global variables from the posterior.

        Parameters
        ----------
        var : str
            Name of the variable to sample.
        num_samples : int
            Number of samples to draw.
        """

        self.predict([var], num_samples=num_samples, idx=self.idx[0:1])

        return self.posterior[var]

    def predict_local_variable(
        self,
        var: str,
        num_samples: int = 25,
        num_split: int = 2048,
        obs_dim: int = 1,
    ):
        """
        Sample local variables from the posterior. In order to
        avoid memory issues, the sampling is performed in batches.

        Parameters
        ----------
        var : str
            Name of the variable to sample.
        num_samples : int
            Number of samples to draw.
        num_split : int
            The parameter determines the size of the batches. The actual
            batch size is total number of observations divided by num_split.
        obs_dim : int
            The dimension of the observations. After sampling, the output
            is concatenated along this dimension.
        """
        split_obs = torch.split(self.idx, num_split)

        # create status bar
        pbar = tqdm(range(len(split_obs)))

        results = []
        for i in pbar:
            self.predict([var], num_samples=num_samples, idx=split_obs[i])
            results.append(self.posterior[var])
            # update status bar
            pbar.set_description(
                f"Predicting {var} for obs {torch.min(split_obs[i])}-{torch.max(split_obs[i])}."
            )

        return np.concatenate(results, obs_dim)



import pyro

SUBSAMPLE = dict(
    num_epochs=5000,
    optimizer=pyro.optim.ClippedAdam,
    optimizer_kwargs={"lr": 0.01, "betas": (0.95, 0.999)},
    scheduler=None,
    loss=pyro.infer.TraceMeanField_ELBO,
    loss_kwargs={"num_particles": 1},
)

DEFAULT = dict(
    num_epochs=20000,
    optimizer=pyro.optim.ClippedAdam,
    optimizer_kwargs={"lr": 0.001, "betas": (0.95, 0.999)},
    scheduler=None,
    loss=pyro.infer.TraceMeanField_ELBO,
    loss_kwargs={"num_particles": 1},
)



from typing import Union

import numpy as np
import pandas as pd
from anndata import AnnData
from scipy.sparse import issparse


def get_rna_counts(adata: AnnData, layers_key: Union[str, None] = None) -> np.ndarray:
    """
    Extracts RNA counts from AnnData object.

    Parameters
    ----------
    adata: AnnData
        AnnData object.
    layers_key: str, optional (default: None)
        If layers_key is None, then the raw counts are extracted from adata.X.
        Otherwise, the counts are extracted from adata.layers[layers_key].

    Returns
    -------
    X: np.ndarray
        RNA counts matrix.
    """
    if layers_key is None:
        X = adata.X
    else:
        X = adata.layers[layers_key]

    if issparse(X):
        X = X.toarray()

    return X.astype(np.float32)


def get_protein_counts(adata: AnnData, protein_obsm_key: str) -> np.ndarray:
    """
    Extracts protein counts from AnnData object.

    Parameters
    ----------
    adata: AnnData
        AnnData object.
    protein_obsm_key: str
        Key for protein counts in adata.obsm.

    Returns
    -------
    Y: np.ndarray
        Protein counts matrix.
    """
    if isinstance(adata.obsm[protein_obsm_key], pd.DataFrame):
        Y = adata.obsm[protein_obsm_key].values
    else:
        Y = adata.obsm[protein_obsm_key]

    return Y.astype(np.float32)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['scCCA', 'scCCA.plots', 'scCCA.train', 'scCCA.utils']

package_data = \
{'': ['*']}

install_requires = \
['pyro-ppl>=1.8.0', 'scanpy>=1.8.2']

extras_require = \
{'docs': ['Sphinx==4.2.0',
          'sphinx-rtd-theme==1.0.0',
          'sphinxcontrib-napoleon==0.7',
          'nbsphinx==0.8.9'],
 'notebook': ['jupyter']}

setup_kwargs = {
    'name': 'sccca',
    'version': '0.1.0',
    'description': 'Single cell canonical correlation analysis.',
    'long_description': None,
    'author': 'Harald Vohringer',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.8.1,<3.11',
}


setup(**setup_kwargs)



import numpy as np
from patsy.design_info import DesignMatrix


def get_states(
    design: DesignMatrix, return_raw: bool = False, indices: bool = True
) -> dict:
    """
    Extracts states and indices of a patsy design matrix.

    Parameters
    ----------
    design: patsy.DesignMatrix
        Patsy design matrix.
    return_raw: bool, optional (default: False)
        If True, then the raw states and combinations are returned.
    indices: bool, optional (default: True)
        If True, then the indices of the states are returned.

    Returns
    -------
    dict
        States and the corresponding indices of the design matrix.
    """
    unique_rows = np.unique(np.asarray(design), axis=0)

    combinations = {}
    for row in range(unique_rows.shape[0]):
        combinations[tuple(np.where(unique_rows[row] == 1)[0])] = unique_rows[row]

    factor_indices = {k.name(): v for k, v in design.design_info.term_slices.items()}
    factor_levels = {
        f.factor.name(): list(f.categories)
        for f in list(design.design_info.factor_infos.values())
    }

    intercept = False
    if "Intercept" in factor_indices.keys():
        intercept = True

    col_idx = {}
    for k, v in factor_indices.items():
        if k != "Intercept":
            for i, j in enumerate(range(v.start, v.stop)):
                # print(k, j)
                if k in factor_levels.keys():
                    col_idx[j] = {k: factor_levels[k][i + 1 if intercept else i]}
                else:
                    col_idx[j] = {k: k}
                # print(col_idx)
        else:
            col_idx[0] = {k: v[0] for k, v in factor_levels.items()}

    states = {}
    for k, v in combinations.items():
        state_dict = {}
        for i in k:
            for f, l in col_idx[i].items():
                state_dict[f] = l

        states[k] = state_dict

    if return_raw:
        return states, combinations

    state_map = {}
    for k, v in states.items():
        c = ""
        for kk, vv in v.items():
            c += f"{vv}|"

        if indices:
            state_map[c.rstrip("|")] = np.where(combinations[k] == 1)[0]
        else:
            state_map[c.rstrip("|")] = combinations[k]

    if len(state_map) == 1 and "" in state_map.keys():
        v = state_map.pop("")
        state_map["base"] = v

    return state_map


def get_state_loadings(adata, model_key: str) -> dict:
    """
    Computes the loading matrix for each state defined in the
    design matrix of the model.

    Parameters
    ----------
    adata: AnnData
        Anndata object with the fitted scPCA model stored.

    model_key: str
        Key of the model in the AnnData object.

    Returns
    -------
    dict of np.ndarray with
        Dictionary with the loading matrices for each state.
    """
    design = adata.uns[model_key]["design"]

    states = {}
    for k, v in design.items():
        states[k] = adata.varm[model_key][..., v].sum(-1)

    return states
